import hashlib
import re
import urllib
import secrets
from typing import List, Optional, Literal, Union
from pydantic import Field

from ..base_model import BaseModel, IgnoredType

from cnc.logger import get_logger

log = get_logger(__name__)


STARTING_PORTS = {
    "POSTGRES": 5432,
    "REDIS": 6379,
    "MYSQL": 3306,
    "SQLSERVER-EX": 1433,
    "SQLSERVER-WEB": 1433,
}


class BaseResourceSettings(BaseModel):
    type: str
    raw_use_existing: Optional[List[dict]] = Field(
        default_factory=list, alias="use_existing"
    )
    engine: Optional[str] = None
    provider: str

    # ------------------------------
    # Parent relationships
    # ------------------------------
    config: Optional[IgnoredType] = IgnoredType()
    service: Optional[IgnoredType] = IgnoredType()

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def collection(self):
        return self.service.environment.collection

    @property
    def application(self):
        return self.service.environment.application

    @property
    def environment(self):
        return self.service.environment

    @property
    def is_resource(self):
        return True

    @property
    def host_output_id(self):
        return f"{self.type}_host_{self.environment.name}_{self.service.instance_name}"

    @property
    def env_var_base(self):
        return f"{self.service.name}".upper().replace("-", "_")

    @property
    def is_web(self):
        return False

    @property
    def is_cache(self):
        return self.type == "cache"

    @property
    def is_filesystem(self):
        return self.type == "filesystem"

    @property
    def is_object_storage(self):
        return self.type == "object_storage"

    @property
    def is_message_queue(self):
        return self.type == "message_queue"

    @property
    def is_database(self):
        return self.type == "database"

    @property
    def unique_id(self):
        hash_str = (
            f"{self.config.environment.application.name}:"
            f"{self.config.environment.collection.name}:"
            f"{self.service.name}:{self.engine}"
        )
        _hash = hashlib.sha256()
        _hash.update(hash_str.encode())
        return _hash.hexdigest()

    @property
    def common_managed_environment_variables(self):
        return {}

    @property
    def use_existing(self):
        return

    @property
    def managed_secret_values_for_tf(self):
        log.warning(f"managed_secret_values_for_tf not defined for {self}")
        return {}

    @property
    def managed_environment_secrets(self):
        return {}

    # ------------------------------
    # Instance methods
    # ------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__} {self.service.name} ({self.type})>"

    __str__ = __repr__


class DatabaseResourceSettings(BaseResourceSettings):
    type: Literal["database"]
    engine: Optional[str] = "postgres"
    snapshot_file_path: Optional[str] = ""
    snapshot_type: Optional[str] = "data"
    raw_adapter: Optional[str] = Field(
        alias="adapter",
        default=None,
    )
    use_db_proxy: Optional[bool] = True
    version: Union[str, int, float]

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def use_existing(self):
        for entry in self.service.environment.existing_resources:
            if entry.name == self.service.name:
                return entry

    @property
    def existing_instance_name(self):
        if self.use_existing:
            return self.use_existing.instance_name

    @property
    def manage_databases(self):
        if self.use_existing:
            return self.use_existing.manage_databases
        return True

    @property
    def database_endpoint(self):
        return self.collection.get_terraform_output(self.host_output_id) or "localhost"

    @property
    def db_name(self):
        if self.use_existing and self.use_existing.db_name:
            return self.use_existing.db_name

        if self.service.environment.application.provider_is_aws:
            dbname = self.service.environment.name.replace("-", "_")
            if re.match("^\\d", dbname):
                return f"db{dbname}"
            else:
                return dbname
        else:
            return self.service.environment.name.lower()

    @property
    def username(self):
        if self.use_existing and self.use_existing.username:
            return self.use_existing.username

        # AWS doesn't like dashes in DB usernames for RDS
        # this is safer across all providers
        return self.service.environment.application.name.replace("-", "_")

    @property
    def connection_string_adapter(self):
        return self.raw_adapter or self.engine.lower()

    @property
    def minor_version_specified(self):
        return bool(isinstance(self.version, float) or float(self.version) % 1)

    @property
    def snapshot_filename(self):
        if self.snapshot_file_path:
            _full = self.snapshot_file_path
            return _full.split("/")[-1]
        return ""

    @property
    def snapshot_bucket_name(self):
        if self.snapshot_file_path:
            _full = self.snapshot_file_path
            return _full.split("/")[0]
        return ""

    @property
    def load_database_snapshot(self):
        if self.snapshot_file_path:
            return True
        return False

    @property
    def managed_environment_variables(self):
        _env = self.common_managed_environment_variables

        _env.update(
            {
                f"{self.env_var_base}_IP": self.database_endpoint,
                f"{self.env_var_base}_HOST": self.database_endpoint,
                f"{self.env_var_base}_ENDPOINT": self.database_endpoint,
                f"{self.env_var_base}_PORT": str(STARTING_PORTS[self.engine.upper()]),
                "DB_NAME": self.db_name,
                "DB_USER": self.username,
                "DB_ENGINE": self.engine.lower(),
            }
        )

        if self.snapshot_file_path:
            _env["SNAPSHOT_FILE_PATH"] = self.snapshot_file_path
            _env["SNAPSHOT_FILE_NAME"] = self.snapshot_filename

        return _env

    @property
    def toolbox_managed_environment_variables(self):
        _env = self.common_managed_environment_variables

        db_password = self.database_password
        db_url = self.database_url(db_password=db_password, local=True)
        _env.update(
            {
                f"{self.env_var_base}_ENDPOINT": "localhost",
                f"{self.env_var_base}_PORT": self.local_port,
                "DB_NAME": self.db_name,
                "DB_USER": self.username,
                "DB_ENGINE": self.engine.lower(),
                f"{self.env_var_base}_DATABASE_URL": db_url,
                "DATABASE_URL": db_url,
                f"{self.env_var_base}_DB_PASSWORD": db_password,
                "DB_PASSWORD": db_password,
            }
        )

        if self.snapshot_file_path:
            _env["SNAPSHOT_FILE_PATH"] = self.snapshot_file_path
            _env["SNAPSHOT_FILE_NAME"] = self.snapshot_filename

        return _env

    @property
    def managed_environment_secrets(self):
        if self.collection.application.provider_is_aws:
            secret_id = f"{self.collection.application.name[:5]}-{self.environment.name}-{self.collection.name}-secrets"
            password_secret_id = f"{secret_id}:{self.service.instance_name}_password"
            url_secret_id = f"{secret_id}:{self.service.instance_name}_url"
        else:
            password_secret_id = "foo"
            url_secret_id = "foo"

        _env = {
            f"{self.env_var_base}_DB_PASSWORD": password_secret_id,
            "DB_PASSWORD": password_secret_id,
            f"{self.env_var_base}_DATABASE_URL": url_secret_id,
            "DATABASE_URL": url_secret_id,
        }

        return _env

    @property
    def cloud_resource_version(self):
        # see https://cloud.google.com/sql/docs/mysql/admin-api/rest/v1beta4/SqlDatabaseVersion
        if self.engine in ["postgres", "mysql"]:
            return f'{self.engine.upper()}_{str(self.version).replace(".", "_")}'

    @property
    def managed_environment_secrets(self):
        return {
            f"{self.env_var_base}_DB_PASSWORD": self.password_secret_id,
            "DB_PASSWORD": self.password_secret_id,
            f"{self.env_var_base}_DATABASE_URL": self.url_secret_id,
            "DATABASE_URL": self.url_secret_id,
        }

    @property
    def managed_secret_values_for_tf(self):
        db_url = self.database_url(self.database_password)
        if "@localhost" in db_url:
            db_url_p1, db_url_p2 = db_url.split("@localhost")
            db_url = f"{db_url_p1}@{self.host_output_tf_value_string}{db_url_p2}"

        return {
            self.url_secret_id.replace("::", "").split(":")[-1]: db_url,
            self.password_secret_id.replace("::", "").split(":")[
                -1
            ]: self.database_password,
        }

    @property
    def database_password(self):
        if not hasattr(self, "_database_password"):
            if self.environment.database_password:
                self._database_password = self.environment.database_password.value
            else:
                self._database_password = (
                    self.password_from_db_secret or secrets.token_hex()[:39]
                )

        return self._database_password

    @property
    def remote_port(self):
        return STARTING_PORTS[self.engine.upper()]

    @property
    def local_port(self):
        starting_port = STARTING_PORTS[self.engine.upper()] + 10
        for i, resource in enumerate(self.environment.database_resources):
            if self.service.instance_name == resource.instance_name:
                return str(starting_port + i)

    # ------------------------------
    # Instance methods
    # ------------------------------

    def database_url(self, db_password="password", local=False):
        if local:
            db_port = self.local_port
            db_endpoint = "localhost"
        else:
            db_port = STARTING_PORTS[self.engine.upper()]
            db_endpoint = self.database_endpoint

        db_url_params = {}
        # SSL is enabled by default for postgres >= 15
        if not self.engine.lower() == "postgres" or int(self.version) < 15:
            db_url_params["sslmode"] = "disable"

        environment = self.service.environment
        if environment.application.provider_is_aws:
            try:
                enc_db_pw = urllib.parse.quote(db_password)
            except Exception as e:
                log.warn(f"Unable to encode db pw for {self.service} | {e}")
                enc_db_pw = db_password

            db_url = (
                f"{self.connection_string_adapter}://{self.username}:"
                f"{enc_db_pw}@{db_endpoint}:{db_port}/{self.db_name}"
            )

            if db_url_params:
                db_url += f"?{urllib.parse.urlencode(db_url_params, safe='/:')}"

            return db_url
        elif environment.application.provider_is_gcp:

            db_url = (
                f"{self.connection_string_adapter}://{self.username}"
                f":{db_password}@{db_endpoint}:{db_port}/{self.db_name}"
            )

            if db_url_params:
                db_url += f"?{urllib.parse.urlencode(db_url_params, safe='/:')}"

            return db_url


class CORSSettings(BaseModel):
    allowed_methods: List[str]
    allowed_origins: List[str]


class CustomBucketHeader(BaseModel):
    name: str
    value: str
    type: Optional[str] = "response"


class BucketResourceSettings(BaseResourceSettings):
    type: Literal["object_storage"]
    cors: Optional[List[CORSSettings]] = []
    custom_headers: Optional[List[CustomBucketHeader]] = []

    @property
    def custom_response_headers(self):
        headers = []
        for header in self.custom_headers:
            if header.type == "response":
                headers.append(header)

        return headers

    @property
    def custom_request_headers(self):
        headers = []
        for header in self.custom_headers:
            if header.type == "request":
                headers.append(header)

        return headers

    @property
    def managed_environment_variables(self):
        _env = {}

        # TODO: make foo real output for bucket name
        _env.update(
            {
                f"{self.service.name.upper()}_URL": self.collection.get_terraform_output(
                    "foo"
                )
                or self.bucket_name,
                f"{self.service.name.upper()}_NAME": self.bucket_name,
            }
        )

        return _env

    @property
    def bucket_name(self):
        if self.use_existing:
            return self.use_existing.instance_name

        name = f"{self.service.name}-{self.config.environment.name}"
        _hash = hashlib.sha256()
        _hash.update(name.encode())
        full_name = (
            f"{self.service.name}-{self.config.environment.name}"
            f"-{_hash.hexdigest()}"
        ).replace("_", "-")[:62]
        # first and last chars must be letters/digits
        first_char = full_name[0]
        if not (first_char.isalpha() or first_char.isdigit()):
            full_name = "b" + full_name[1:]

        last_char = full_name[-1]
        if not (last_char.isalpha() or last_char.isdigit()):
            full_name = full_name[:61] + "b"
        return full_name


class FilesystemResourceSettings(BaseResourceSettings):
    type: Literal["filesystem"]
    mountpoint: Optional[str] = None

    @property
    def existing_filesystem_id(self):
        if self.use_existing and self.is_filesystem:
            if self.service.environment.project_type == self.use_existing.project_type:
                return self.use_existing.instance_name

    @property
    def existing_filesystem_mountpoint(self):
        if self.use_existing and self.is_filesystem:
            if self.service.environment.project_type == self.use_existing.project_type:
                return self.mountpoint or self.use_existing.instance_name


class QueueResourceSettings(BaseResourceSettings):
    type: Literal["message_queue"]
    fifo: Optional[bool] = True

    @property
    def managed_environment_variables(self):
        _env = self.common_managed_environment_variables
        # TODO: this wont work - need to make sure url is set properly here
        _env.update(
            {
                f"{self.name.upper()}_URL": self.service.environment.data.get(
                    f"sqs_queue_{self.queue_name}"
                ),
                f"{self.name.upper()}_NAME": self.queue_name,
            }
        )
        return _env

    @property
    def use_existing(self):
        for entry in self.service.environment.existing_resources:
            if entry.name == self.service.name:
                return entry

    @property
    def tf_safe_queue_name(self):
        # cannot include .fifo at the end
        return self.queue_name.replace(".fifo", "")

    @property
    def queue_name(self):
        # Queue names must be unique for your AWS account and region
        # must be made up of only uppercase and lowercase ASCII letters,
        # numbers, underscores, and hyphens, and must be between 1 and 80 characters long.
        # For a FIFO (first-in-first-out) queue, the name must end with the .fifo suffix.
        if self.use_existing and self.use_existing.instance_name:
            return self.use_existing.instance_name

        queue_name = f"{self.environment.instance_name[:37]}-{self.name[-37:]}"
        if self.is_fifo:
            queue_name += ".fifo"

        return queue_name


class CacheResourceSettings(BaseResourceSettings):
    type: Literal["cache"]
    engine: Optional[str] = "redis"
    version: Union[str, int, float] = "6"

    @property
    def redis_ip_output_id(self):
        return f"redis_ip_{self.service.instance_name}"

    @property
    def redis_port_output_id(self):
        return f"redis_port_{self.service.instance_name}"

    @property
    def redis_ip(self):
        return (
            self.collection.get_terraform_output(self.redis_ip_output_id) or "localhost"
        )

    @property
    def redis_port(self):
        return str(
            self.collection.get_terraform_output(self.redis_port_output_id)
            or STARTING_PORTS[self.engine.upper()]
        )

    @property
    def managed_environment_variables(self):
        _env = self.common_managed_environment_variables
        _env.update(
            {
                f"{self.env_var_base}_PORT": self.redis_port,
                f"{self.env_var_base}_IP": self.redis_ip,
                f"{self.env_var_base}_HOST": self.redis_ip,
                f"{self.env_var_base}_URL": f"redis://{self.redis_ip}:{self.redis_port}",
                "REDIS_URL": f"redis://{self.redis_ip}:{self.redis_port}",
            }
        )
        return _env

    @property
    def toolbox_managed_environment_variables(self):
        _env = self.common_managed_environment_variables
        _env.update(
            {
                f"{self.env_var_base}_PORT": self.local_port,
                f"{self.env_var_base}_IP": "host.docker.internal",
                f"{self.env_var_base}_URL": f"redis://host.docker.internal:{self.local_port}",
                "REDIS_URL": f"redis://host.docker.internal:{self.local_port}",
            }
        )
        return _env

    @property
    def is_cache(self):
        return True

    @property
    def local_port(self):
        starting_port = STARTING_PORTS[self.engine.upper()]
        for i, resource in enumerate(self.environment.cache_resources):
            if self.service.instance_name == resource.instance_name:
                return str(starting_port + i)



from pydantic import (
    BaseModel as PydanticBaseModel,
    model_validator,
    ValidationInfo,
)

from cnc.logger import get_logger

log = get_logger(__name__)


class IgnoredType:
    pass


class BaseModel(PydanticBaseModel):
    @model_validator(mode="before")
    @classmethod
    def ensure_provider(cls, data: dict, info: ValidationInfo):
        if info.context:
            data["provider"] = info.context.get("provider")

        return data

    @classmethod
    def convert_dict_keys_to_names(cls, value: dict) -> list[dict]:
        with_names = []

        for k, v in value.items():
            if not isinstance(v, dict):
                raise ValueError(f"{k} is not a dict, bad data format")

            v.update({"name": k})
            with_names.append(v)

        return with_names

    class Config:
        ignored_types = (IgnoredType,)
        arbitrary_types_allowed = True



from .base_test_class import CNCBaseTestCase
from cnc.models import Application

from cnc.logger import get_logger

log = get_logger(__name__)


class EnvironmentCollectionBaseTestCase(CNCBaseTestCase):
    fixture_name = "backend-1-service-1-db"
    env_data_filepath = "environments.yml"

    def setUp(self):
        super().setUp()
        app = Application.from_environments_yml(self.env_data_filepath)
        self.collection = app.collections[0]


class EnvironmentCollectionTestCase(EnvironmentCollectionBaseTestCase):

    def test_collection_env_by_name(self):
        self.assertIsNone(self.collection.environment_by_name("foo"))
        self.assertIsNotNone(self.collection.environment_by_name("main"))


class EnvironmentCollectionExistingDBTestCase(EnvironmentCollectionBaseTestCase):
    fixture_name = "backend-1-service-1-db-use-existing"

    def test_collection_use_existing_db(self):
        self.assertEqual(
            self.collection.environments[0].existing_resources[0].name, "db1"
        )
        self.assertEqual(
            self.collection.environments[0].existing_resources[0].instance_name,
            "foobar",
        )


class EnvironmentCollectionTwoServicesandDBsTwoEnvsTest(
    EnvironmentCollectionBaseTestCase
):
    fixture_name = "backend-1-service-2-db-2-envs"

    def test_environment_order(self):
        self.assertEqual(len(self.collection.environments), 2)
        self.assertEqual(self.collection.environments[0].name, "main")
        self.assertEqual(self.collection.environments[1].name, "staging")

    def test_environment_services(self):
        self.assertEqual(len(self.collection.all_services), 6)
        self.assertEqual(len(self.collection.all_services_for_type("backend")), 2)
        self.assertEqual(len(self.collection.all_services_for_type("database")), 4)
        self.assertEqual(len(self.collection.all_services_for_type("frontend")), 0)
        self.assertEqual(len(self.collection.all_services_for_type("cache")), 0)


class EnvironmentCollectionRegionSettings(EnvironmentCollectionBaseTestCase):
    fixture_name = "backend-1-service-2-db-2-envs"
    env_data_filepath = "environments_gcp_run_region_settings.yml"

    def test_region_settings(self):
        self.assertEqual(self.collection.region, self.collection.collection_region)
        self.assertEqual(self.collection.region, "us-west2")
        self.assertEqual(self.collection.application.region, "us-east1")



from typing import Literal
from cnc.models.config.resource import DatabaseResourceSettings
from cnc.logger import get_logger

log = get_logger(__name__)


class GCPDatabaseResourceSettings(DatabaseResourceSettings):
    provider: Literal["gcp"]

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def password_secret_id(self):
        return f"{self.service.instance_name}_password"

    @property
    def url_secret_id(self):
        return f"{self.service.instance_name}_url"

    @property
    def host_output_tf_value_string(self):
        tf_resource_type = "google_sql_database_instance."
        host_attr = ".private_ip_address"

        if "lite" in self.application.flavor:
            host_attr = ".public_ip_address"

        if self.use_existing:
            tf_resource_type = f"data.{tf_resource_type}"

        return "".join(
            ["${", tf_resource_type, self.service.instance_name, host_attr, "}"]
        )

    @property
    def password_from_db_secret(self):
        try:
            return self.collection.get_secret_value(self.password_secret_id)
        except Exception as e:
            log.debug(f"No existing db secret found for {self.service}: {e}")
            return None

    @property
    def toolbox_managed_environment_variables(self):
        if "lite" in self.application.flavor:
            _env = self.managed_environment_variables

            db_password = self.database_password
            db_url = self.database_url(db_password=db_password)
            _env.update(
                {
                    f"{self.env_var_base}_DATABASE_URL": db_url,
                    "DATABASE_URL": db_url,
                    f"{self.env_var_base}_DB_PASSWORD": db_password,
                    "DB_PASSWORD": db_password,
                }
            )
            return _env

        return super().toolbox_managed_environment_variables

    @property
    def toolbox_ssh_port_mapping(self):
        return f"{self.local_port}:{self.database_endpoint}:{self.remote_port}"



from tabulate import tabulate
import typer

from .telemetry import send_event

from cnc.logger import get_logger

log = get_logger(__name__)


app = typer.Typer()


@app.command("app")
def application_info(
    ctx: typer.Context,
):
    send_event("info.app")
    app = ctx.obj.application

    data = {
        "name": app.name,
        "version": f"{app.provider}/{app.flavor}/{app.version}",
        "collections": ",".join([c.name for c in app.collections]),
    }

    headers = list(data.keys())
    table_data = [list(data.values())]

    print(tabulate(table_data, headers, tablefmt="grid"))

    raise typer.Exit()


@app.command()
def environments(
    ctx: typer.Context,
    collection_name: str = "",
):
    send_event("info.environments")
    collection = (
        ctx.obj.application.collection_by_name(collection_name)
        or ctx.obj.application.default_collection
    )
    if not collection:
        log.error(f"No collection found for: {collection_name}")
        raise typer.Exit(code=1)

    env_datas = []
    for environment in collection.environments:
        env_datas.append(environment.collection.cli_info_for_environment(environment))

    if env_datas:
        headers = list(env_datas[0].keys())
        table_data = [list(d.values()) for d in env_datas]
        print(tabulate(table_data, headers, tablefmt="grid"))
    else:
        print(f"No environments for {collection}")

    raise typer.Exit()


@app.command()
def services(
    ctx: typer.Context,
    environment_name: str,
    collection_name: str = "",
):
    send_event("info.services")
    collection = (
        ctx.obj.application.collection_by_name(collection_name)
        or ctx.obj.application.default_collection
    )
    if not collection:
        log.error(f"No collection found for: {collection_name}")
        raise typer.Exit(code=1)

    environment = collection.environment_by_name(environment_name)
    if not environment:
        log.error(f"No environment found for: {environment_name}")
        raise typer.Exit(code=1)

    svc_datas = []
    for service in environment.services:
        svc_datas.append(service.cli_info())

    if svc_datas:
        headers = list(svc_datas[0].keys())
        table_data = [list(d.values()) for d in svc_datas]
        print(tabulate(table_data, headers, tablefmt="grid"))
    else:
        print(f"No services for {environment}")

    raise typer.Exit()



import subprocess
from typing import List

from .cycle_stage_base import EnvironmentTemplatedBase
from cnc.logger import get_logger

log = get_logger(__name__)


class ToolboxManager(EnvironmentTemplatedBase):
    template_type = "toolbox"

    def __init__(self, service, service_tags=None):
        super().__init__(service.environment, service_tags)
        self.service = service

    @property
    def environment_items(self):
        if not hasattr(self, "_environment_items"):
            all_env_items = self.service.environment_items
            # Append toolbox env vars
            for resource in self.environment.config.resources:
                if not hasattr(
                    resource.settings,
                    "toolbox_managed_environment_variables",
                ):
                    continue

                for (
                    name,
                    value,
                ) in resource.settings.toolbox_managed_environment_variables.items():
                    # Remove existing var if it is being replaced
                    for i, env_item in enumerate(all_env_items):
                        if env_item.name == name:
                            del all_env_items[i]

                    all_env_items.append(
                        self.environment.config.variable_object(
                            {"name": name, "value": value}
                        )
                    )

            self._environment_items = all_env_items

        return self._environment_items

    def template_context(self, _service=None, command=None):
        return {
            "toolbox": self,
            "service": self.service,
            "environment": self.environment,
            "collection": self.collection,
            "bastion_instance_id": self.collection.get_terraform_output(
                "bastion_instance_id"
            ),
            "command": command,
        }

    def render_toolbox(self, command=None):
        context = self.template_context(command=command)
        context["render_template"] = self.write_template_with_context(self.service)
        self.write_template("main.sh.j2", context=context)

    def copy_templates(self):
        self.copy_template_dir()

    def start(self):
        log.debug(f"Rendering toolbox script for {self} @ {self.config_files_path}")
        self.setup()
        self.render_toolbox()
        log.debug(f"Done rendering toolbox script for {self}, starting toolbox...")

        try:
            ret = subprocess.run(
                [
                    "bash",
                    "-c",
                    (
                        f"chmod +x {self.rendered_files_path}/main.sh "
                        f"&& setsid {self.rendered_files_path}/main.sh"
                    ),
                ],
                executable="/bin/bash",
            )
            if ret.stdout or ret.stderr:
                log.debug(f"Output from main.sh: \n{ret.stdout}\n{ret.stderr}")
        except subprocess.CalledProcessError as e:
            # Handle any errors that occurred during script execution
            print(f"Error running the script: {e}")
            print("Error output:")
            print(e.stderr)

    def run(self, command: List[str] = []):
        log.debug(f"Rendering toolbox script for {self} @ {self.config_files_path}")
        self.setup()
        self.render_toolbox(command=" ".join(command))
        log.debug(f"Done rendering toolbox script for {self}, starting toolbox...")

        try:
            ret = subprocess.run(
                [
                    "bash",
                    "-c",
                    (
                        f"chmod +x {self.rendered_files_path}/main.sh "
                        f"&& setsid {self.rendered_files_path}/main.sh"
                    ),
                ],
                executable="/bin/bash",
            )
            if ret.stdout or ret.stderr:
                log.debug(f"Output from main.sh: \n{ret.stdout}\n{ret.stderr}")

            return ret.returncode
        except subprocess.CalledProcessError as e:
            # Handle any errors that occurred during script execution
            print(f"Error running the script: {e}")
            print("Error output:")
            print(e.stderr)



import os
import shutil
import boto3
import json
from cnc.models import EnvironmentCollection
from typing import Literal
from cnc.logger import get_logger

log = get_logger(__name__)


class AWSEnvironmentCollection(EnvironmentCollection):
    provider: Literal["aws"]

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def hosted_zone_ns_records(self):
        return self.get_terraform_output("hosted_zone_ns_records")

    @property
    def task_execution_role_arn(self):
        return f"arn:aws:iam::{self.account_id}:role/" f"{self.instance_name}-ecs-task"

    @property
    def existing_vpc_config(self):
        for resource in self.existing_resources:
            if resource.name.lower() == "vpc":
                return resource

    @property
    def vpc_resource_address(self):
        if self.existing_vpc_config:
            return f"data.aws_vpc.{self.instance_name}"

        return f"aws_vpc.{self.instance_name}"

    # ------------------------------
    # Instance methods
    # ------------------------------

    def cli_info_for_environment(self, environment):
        return {
            "name": environment.name,
            "hosted_zone_ns_records": self.get_terraform_output(
                "hosted_zone_ns_records"
            ),
            "domains": environment.domains,
        }

    def preconfigure_infrastructure(self, should_cleanup=True):
        config = self.provision_stage_manager
        is_ok = config.make_ready_for_use(should_cleanup=should_cleanup)

        results = []
        steps = []

        if not is_ok:
            log.info(f"Cannot init TF for {self}")
            return {"ok": False, "results": results, "steps": steps}

        total_plan_changes = config.plan(save=True, plan_filename="total")

        if not total_plan_changes:
            log.debug(
                "Did not get any total plan changes! Calling apply and logging..."
            )
            try:
                _ret = config.apply()
            except Exception as e:
                log.debug(f"Error in calling apply... {e}")
                _ret = {"@message": f"Error: {e}"}
            results.append(_ret)

            log.debug(f"Exiting infra configure step for {self}")
            return {"ok": False, "results": results, "steps": steps}

        if total_plan_changes["status"] == "locked":
            steps.append(total_plan_changes)
            log.warning(
                (
                    "TF is locked (call config.unlock(LOCK_ID) in a toolbox if"
                    f" safe) for {self}: {total_plan_changes}"
                )
            )
            return {"ok": False, "results": results, "steps": steps}

        total_plan_changes["resource_changes"] = {
            "create": [],
            "update": [],
            "delete": [],
        }
        plan_json = config.plan_json(plan_filename="total") or {}
        for rc in plan_json.get("resource_changes", []):
            actions = rc["change"]["actions"]
            resource_data = {key: rc[key] for key in rc.keys() if key not in ["change"]}
            resource_data["actions"] = actions
            if "create" in actions:
                total_plan_changes["resource_changes"]["create"].append(resource_data)
            elif "delete" in actions:
                total_plan_changes["resource_changes"]["delete"].append(resource_data)
            elif "update" in actions:
                total_plan_changes["resource_changes"]["update"].append(resource_data)

        steps.append(total_plan_changes)

        return {"ok": True, "results": results, "steps": steps}

    def configure_infrastructure(self, should_cleanup=True):
        config = self.provision_stage_manager

        if not config.saved_plan_exists(plan_filename="total"):
            res = self.preconfigure(should_cleanup=should_cleanup)
            if not res["ok"]:
                return res
        else:
            res = {"steps": [], "results": []}
            res["steps"].append(config.plan_json(plan_filename="total"))

        _ret = config.apply(use_saved=True, plan_filename="total")
        log.debug(f"TF apply output: {_ret}")

        _ok = True
        if _ret.get("@message", "").startswith("Error: "):
            _ok = False

        res["results"].append(_ret)

        config.cleanup()
        return {"ok": _ok, "results": res["results"], "steps": res["steps"]}

    def get_secret_value(self, secret_id, version_id=None):
        secrets_client = boto3.client("secretsmanager")

        inner_json_key = None
        secret_id_parts = secret_id.split(":")
        if len(secret_id_parts) > 1:
            secret_id, inner_json_key = secret_id_parts

        kwargs = {"SecretId": secret_id}
        if version_id:
            kwargs["VersionId"] = version_id

        secret_string = secrets_client.get_secret_value(**kwargs).get("SecretString")
        if inner_json_key:
            return json.loads(secret_string)[inner_json_key]

        return secret_string

    def generate_tf_assets(self, config_files_path, rendered_files_path):
        log.debug(f"Generating provider assets for {self}...")
        # Zip the js file
        # ensure its in env_collection config_files_path
        zip_file_path = os.path.join(rendered_files_path, "frontend_routing_lambda")
        shutil.make_archive(
            zip_file_path,
            "zip",
            root_dir=(f"{config_files_path}/frontend_routing_lambda"),
        )

        return True



import typer
import time
from typing import List
from typing_extensions import Annotated

from cnc.models import DeployStageManager
from .telemetry import send_event

from cnc.logger import get_logger

log = get_logger(__name__)


app = typer.Typer()


@app.command()
def perform(
    ctx: typer.Context,
    environment_name: str,
    service_tags: List[str] = typer.Option(
        [],
        "--service-tag",
        "-t",
        help="Set the tag to use for this service with svc_name=tag, default is 'int(time.time())'. If any provided, only builds provided services and will build all services if empty",
    ),
    default_tag: Annotated[
        str,
        typer.Option(
            "-d", "--default-tag", envvar="CNC_DEFAULT_TAG", help="CNC default tag"
        ),
    ] = None,
    collection_name: str = "",
    cleanup: bool = True,
    debug: bool = False,
    generate: bool = True,
):
    start_time = time.time()
    send_event("deploy.perform")
    collection = ctx.obj.application.collection_by_name(collection_name)
    if not collection:
        log.error(f"No collection found for: {collection_name}")
        raise typer.Exit(code=1)

    environment = collection.environment_by_name(environment_name)
    if not environment:
        log.error(f"No environment found for: {environment_name}")
        raise typer.Exit(code=1)

    deployer = DeployStageManager(
        environment,
        service_tags=service_tags,
        default_tag=default_tag,
    )
    deployer.perform(
        should_cleanup=cleanup,
        should_regenerate_config=generate,
        debug=debug,
    )

    log.debug(
        f"All set deploying for {deployer.rendered_files_path} in {int(time.time() - start_time)} seconds"
    )
    raise typer.Exit()
